/* avl.c */
/* Algoritmos para rvores AVL (Wirth) */

#include <stdio.h>
#include <stdlib.h>
#include "avl.h"
#include "funcoes.h"

avl::avl() {
  raiz = NULL;
  desbalanceado = F;
  achou = F;
}

celula_pt avl::avl_busca(info x) {
  celula_pt pt = raiz;

  while (pt != NULL && !compara_igual_chave(x, pt)) // && (pt->i.chave != x.chave))
    if (compara_menor_chave(x, pt)) // x.chave < pt->i.chave)
      pt = pt->esq;
    else // no precisa dessa comparao if (x.chave > pt->i.chave)
      pt = pt->dir;
  return pt;
}

/* Funes pblicas de insero e remoo */
void avl::avl_insere(info x) {
  real_insere(&raiz, x);
}

void avl::avl_delete(info x) {
  real_delete(&raiz, x);
}

/* Insercao em Arvore AVL */ 
void avl::real_insere(celula_pt *pai, info x)
{  
  if (!(*pai)) /* Insere */
  {
    achou = F;
    desbalanceado = V;
    if ((*pai = new celula) == NULL)
    {
      printf("Erro: memoria insuficiente!\n");
      exit(0);
    }    
    (*pai)->esq = (*pai)->dir = NULL;
    (*pai)->bal = 0;
    (*pai)->i = x;
    (*pai)->pai = NULL;
  }    
  else if (compara_menor_chave(x, (*pai)) || compara_igual_chave(x, (*pai))) // (x.chave <= (*pai)->i.chave)
  {
    real_insere(&(*pai)->esq, x);
    if (desbalanceado) /* se cresceu o ramo esquerdo */
      switch((*pai)->bal)
      {
      case -1 : (*pai)->bal = 0;	
	(*pai)->esq->pai = (*pai);
	desbalanceado = F;
	break;
      case  0 : (*pai)->bal = 1;
	(*pai)->esq->pai = (*pai);
	break;
      case  1 : rotacao_esq(pai);
	desbalanceado = F;
      }
  }
  
  else if (!compara_menor_chave(x, (*pai)) && !compara_igual_chave(x, (*pai)))//(x.chave > (*pai)->i.chave)
  {
    real_insere(&(*pai)->dir, x);
    if (desbalanceado) /* se cresceu o ramo direito */
      switch((*pai)->bal)
      {
      case  1 : (*pai)->bal = 0;
	(*pai)->dir->pai = (*pai);
	desbalanceado = F;
	break;
      case  0 : (*pai)->bal = -1;
	(*pai)->dir->pai = (*pai);
	break;
      case -1 : rotacao_dir(pai);
	desbalanceado = F;
      }
  }
  else
  {
    achou = V;
    desbalanceado = F;    
    fprintf(stderr, "ERRO - RealInsere(): A chave j est na rvore\n");
    exit(1);
  }
}

/* rotaciona a arvore para a esquerda */
void avl::rotacao_esq (celula_pt *pai)
{
  celula_pt neto, filho;
  filho = (*pai)->esq;
  if (filho->bal >= 0) 
  {
    (*pai)->esq = filho->dir;
    filho->dir = *pai;

    // novo
    filho->pai = (*pai)->pai;    
    (*pai)->pai = filho;
    if ((*pai)->esq)
      (*pai)->esq->pai = (*pai);
    // novo

    if (filho->bal == 0) {
	 (*pai)->bal = 1;
	 filho->bal = -1;
	 desbalanceado = F;
    }
    else { 
	 (*pai)->bal = 0;
	 filho->bal = 0;
	 desbalanceado = V;
    }
    (*pai) = filho;
  }  
  else
  {
    desbalanceado = V;
    neto = filho->dir;
    filho->dir = neto->esq;
    neto->esq = filho;
    (*pai)->esq = neto->dir;
    neto->dir = *pai;

    // novo
    neto->pai = (*pai)->pai;
    filho->pai = neto;
    (*pai)->pai = neto;
    if (filho->dir)
      filho->dir->pai = filho;
    if ((*pai)->esq)
      (*pai)->esq->pai = (*pai);
    // novo

    switch(neto->bal)
    {
    case  1 : (*pai)->bal = -1;
      filho->bal = 0;
      break;
    case  0 : (*pai)->bal = filho->bal = 0;
      break;
    case -1 : (*pai)->bal = 0;
      filho->bal = 1;
    }
    neto->bal = 0;
    (*pai) = neto;
  }
}

/* rotaciona a arvore para a direita */
void avl::rotacao_dir (celula_pt *pai)
{
  celula_pt neto, filho;
  filho = (*pai)->dir;
  if (filho->bal <= 0)
  {
    (*pai)->dir = filho->esq;
    filho->esq = *pai;

    // novo
    filho->pai = (*pai)->pai;    
    (*pai)->pai = filho;
    if ((*pai)->dir)
      (*pai)->dir->pai = (*pai);
    // novo


    if (filho->bal == 0) {
	 (*pai)->bal = -1;
	 filho->bal = 1;
	 desbalanceado = F;
    }
    else {	 
	 (*pai)->bal = 0;
	 filho->bal = 0;
	 desbalanceado = V;
    }
    (*pai) = filho;
  }  
  else 
  {
    desbalanceado = V;
    neto = filho->esq;
    filho->esq = neto->dir;
    neto->dir = filho;
    (*pai)->dir = neto->esq;
    neto->esq = *pai;

    // novo
    neto->pai = (*pai)->pai;
    filho->pai = neto;
    (*pai)->pai = neto;
    if (filho->esq)
      filho->esq->pai = filho;
    if ((*pai)->dir)
      (*pai)->dir->pai = (*pai);
    // novo

    switch(neto->bal)
    {
    case -1 : (*pai)->bal = 1;
      filho->bal = 0;
      break;
    case 0 : (*pai)->bal = filho->bal = 0;
      break;
    case 1 : (*pai)->bal = 0;
      filho->bal = -1;
    }
    neto->bal = 0;
    *pai = neto;
  }
}

/* ajusta o balanco para a direita quando remove alguma chave */
void avl::balanco_dir (celula_pt *p) {
  switch((*p)->bal)
  {
  case -1: (*p)->bal = 0;
    break;
  case 0: (*p)->bal = 1;
    desbalanceado = F;
    break;
  case 1:
    rotacao_esq(p);
  }
}

/* arruma o balanco para a esquerda quando remove alguma chave */
void avl::balanco_esq (celula_pt *p) {
  switch((*p)->bal)
  {
  case 1: (*p)->bal = 0;
    break;
  case 0: (*p)->bal = -1;
    desbalanceado = F;
    break;
  case -1:
    rotacao_dir(p);
  }
}

/* funcao auxiliar para a Delete */
void avl::delete_aux(celula_pt *q, celula_pt *r) {
  if ((*r)->dir != NULL) {
    delete_aux(q, &(*r)->dir);
    if (desbalanceado)
      balanco_dir(r);
  }
  else {
    (*q)->i.chave = (*r)->i.chave;
    //(*q)->bal = (*r)->bal;
    (*r)->bal = (*q)->bal;
    if ((*r)->esq)
      (*r)->esq->pai = (*q)->pai;
    (*q) = (*r);
    *r = (*r)->esq;
    desbalanceado = V;
  }
}

/* ATENO: se existirem chaves repetidas remove a que estiver mais  esquerda */
void avl::real_delete(celula_pt *pai, info x) {
  celula_pt q;
  
  if ((*pai) != NULL) {
    if (compara_menor_chave(x, (*pai))) {//((*pai)->i.chave > x.chave) {
      real_delete(&(*pai)->esq, x);
      if (desbalanceado)
	balanco_esq(pai);
    }
    else if (!compara_igual_chave(x, (*pai))) {//((*pai)->i.chave < x.chave) {      
      real_delete(&(*pai)->dir, x);
      if (desbalanceado)
	balanco_dir(pai);
    }
    else if (compara_igual_chave(x, (*pai)) && (*pai)->esq != NULL && compara_igual_chave(x, (*pai)->esq)) {
      //((*pai)->i.chave == x.chave && (*pai)->esq != NULL && (*pai)->esq->i.chave == x.chave) {
      real_delete(&(*pai)->esq, x);
      if (desbalanceado) {
	balanco_esq(pai);
      }
    }
    else {
      q = (*pai);
      if (q->dir == NULL) {
	if (q->esq)
	  q->esq->pai = (*pai)->pai;
	(*pai) = q->esq;
	desbalanceado = V;
      }
      else if (q->esq == NULL) {
	if (q->dir)
	  q->dir->pai = (*pai)->pai;
	(*pai) = q->dir;
	desbalanceado = V;
      }
      else {
	delete_aux(&q, &q->esq);
	if (desbalanceado)
	  balanco_esq(pai);	
      }
      //free(q);
      //delete q;
    }
  }
  else {
    desbalanceado = F;
    fprintf(stderr, "RealDelete(): ERRO - Esta chave no est na rvore\n");
    exit(1);
  }
}


/*int avl::compara_igual_chave (info a, celula_pt b) {
  switch(tc) {
  case (FILA_EVENTO): 
    return (a.chave.chave_d == b->i.chave.chave_d);
  case (LINHA_VARREDURA):
    return (int) Collinear(&(reta(b)->a), &(reta(b)->b), &(a.p->a));
  }
  return 0; 
}

int avl::compara_menor_chave (info a, celula_pt b) {
  switch(tc) {
  case (FILA_EVENTO):
    return (a.chave.chave_d < b->i.chave.chave_d);
  case (LINHA_VARREDURA):
    return (int) !LeftOn(&(reta(b)->a), &(reta(b)->b), &(a.p->a));
  }
  return 0; 
  }*/

celula_pt avl::avl_minimo () {
  celula_pt p, ant;
  
  p = raiz;
  ant = nil;
  while (p != nil) {
    ant = p;
    p = p->esq;
  }
  return ant;
}

celula_pt avl::avl_maximo () {
  celula_pt p, ant;
  
  p = raiz;
  ant = nil;
  while (p != nil) {
    ant = p;
    p = p->dir;
  }
  return ant;
}

void avl::imprime_arvore() {
  printf("---> IMPRIMINDO RVORE <---\n");
  imprime_aux(raiz, 0);
}

void avl::imprime_aux(celula_pt pai, int espaco) {
  if (!pai) return;
  for (int i = 0; i < espaco; i++) printf(" ");      
  if (pai->pai) {
    if (pai->pai->dir == pai)
      printf("Direita %d %p %p %p %p %p %p: ", espaco, pai->pai, pai->pai->dir, pai->pai->esq, pai, pai->dir, pai->esq);
    else if (pai->pai->esq == pai)
      printf("Esquerda %d %p %p %p %p %p %p: ", espaco, pai->pai, pai->pai->dir, pai->pai->esq, pai, pai->dir, pai->esq);
    else
      printf("FUDEU! Esq? Dir? %p %p %p %p: ", pai->pai, pai->pai->dir, pai->pai->esq, pai);
  }
  else printf("Raiz %p %p: ", pai, pai->pai?pai->pai:pai);
  printf("(%.1f, %.1f), (%.1f, %.1f)\n", reta(pai)->a.x, reta(pai)->a.y, reta(pai)->b.x, reta(pai)->b.y);
  //printf("(%.1f, %.1f)\n", pai->i.p->a.x, pai->i.p->a.y);
  imprime_aux(pai->dir, espaco+1);
  imprime_aux(pai->esq, espaco+1);
}

int avl::vazia() {
  return (raiz == NULL);
}

int avl::compara_igual_chave (info a, celula_pt b) {
  return (a.chave.chave_d == b->i.chave.chave_d);
}

int avl::compara_menor_chave (info a, celula_pt b) {
  return (a.chave.chave_d < b->i.chave.chave_d);
}
