/****************************************************************************
 * funcoes.c                                                                *
 * Neste arquivo implementamos algumas funes genricas bsicas            *
 ****************************************************************************/

#include "funcoes.h"

double maiory, menory;

double Area2(ponto * a, ponto * b, ponto * c) {
  return a->x * b->y - a->y * b->x + a->y * c->x - a->x * c->y + b->x * c->y
	- c->x * b->y;
}

double Left(ponto * a, ponto * b, ponto * c) {
  return (Area2(a, b, c) > SMALL);
}

double LeftOn(ponto * a, ponto * b, ponto * c) {
  return (Area2(a, b, c) >= -SMALL);
}

double Collinear(ponto * a, ponto * b, ponto * c) {
  double aux = Area2(a, b, c);
  return (aux > -SMALL && aux < SMALL);
}


/* Esta funo verifica se duas retas (a, b) e (c, d) tem interseco */
int interseccao_reta(reta * r1, reta * r2) {
    /* Vamos acertar os lados dos nossos segmentos para podermos fazer os
     * testes rpidos */
    ponto * a, * b, * c, * d;
    double axT, ayT, bxT, byT, cxT, cyT, dxT, dyT;
    a = & r1->a;
    b = & r1->b;
    c = & r2->a;
    d = & r2->b;

    if (a->x <= b->x) {
	axT = a->x;
	bxT = b->x;
    }
    else {
	axT = b->x;
	bxT = a->x;
    }

    if (a->y <= b->y) {
	ayT = a->y;
	byT = b->y;
    }
    else {
	ayT = b->y;
	byT = a->y;
    }

    if (c->x <= d->x) {
	cxT = c->x;
	dxT = d->x;
    }
    else {
	cxT = d->x;
	dxT = c->x;
    }

    if (c->y <= d->y) {
	cyT = c->y;
	dyT = d->y;
    }
    else {
	cyT = d->y;
	dyT = c->y;
    }

    /* Teste rpido */
    if (!((axT <= dxT) && (cxT <= bxT) && (ayT <= dyT) && (cyT <= byT)))
	return 0; /* Nosso teste rpido deu pau */

    /* Agora os testes difceis */
    if ((Left(a, b, c) != Left(a, b, d)) && (Left(c, d, a) != Left(c, d, b)))
	return 1;
    
    if (Collinear(a, b, c) || Collinear(a, b, d) || Collinear(c, d, a) ||
	Collinear(c, d, b))
	return 1;

    return 0;
}

/* Esta funo calcula o ponto de interseco entre duas retas e devolve 0
 * caso no exista interseco */
ponto * ponto_interseccao_reta(reta * r1, reta * r2, ponto * p) {
    /* Vamos acertar os lados dos nossos segmentos para podermos fazer os
     * testes rpidos */
    ponto * a, * b, * c, * d;
    double s, t;
    double num, denom;
    a = & r1->a;
    b = & r1->b;
    c = & r2->a;
    d = & r2->b;

    denom = a->x * (d->y - c->y) + b->x * (c->y - d->y) + d->x * (b->y - a->y)
	+ c->x * (a->y - b->y);

    /* Verificamos se os segmentos so paralelos */
    if (denom == 0) {

      /* precisamos verificar se as retas paralelas se intersectam */
      /* este pedao precisa de otimizao */
      if (((a->x > c->x) && (a->x > d->x) && (b->x > c->x) && (b->x > d->x))
	  || ((a->x < c->x) && (a->x < d->x) && (b->x < c->x) && (b->x <
								    d->x)))
	    return 0;
	if (((a->y > c->y) && (a->y > d->y) && (b->y > c->y) && (b->y > d->y))
	    || ((a->y < c->y) && (a->y < d->y) && (b->y < c->y) && (b->y <
								    d->y)))
	    return 0;
	if (Collinear(a, b, c) || Collinear(a, b, d) || Collinear(a, c, d) ||
	    Collinear(b, c, d)) {
	  if (((a->x >= c->x) && (a->x <= d->y)) || ((a->x <= c->x) && (a->x
									>=
									d->y))) {
                p->x = a->x;
		p->y = a->y;
	    }
	    else {
		p->x = b->x;
		p->y = b->y;
	    }

	    return p;
	}
	return 0;
    }

    num = a->x * (d->y - c->y) + c->x * (a->y - d->y) + d->x * (c->y - a->y);

    s = num / denom;

    num = - (a->x * (c->y - b->y) + b->x * (a->y - c->y) + c->x * (b->y -
								   a->y));

    t = num / denom;

    p->x = a->x + s * (b->x - a->x);
    p->y = a->y + s * (b->y - a->y);

    if ((0 <= s) && (s <= 1) && (0 <= t) && (t <= 1))
	return p;
    
    return 0;
}



/* Esta funo l as retas de entrada e monta uma lista ligada de retas e
 * devolve o nmero de elementos lidos */
int le_pontos(FILE * Entrada, reta * retas) {
  double ax, ay, bx, by; 
  reta * atual = 0;
  int contador = 0;
  int inicio = 1;
  
  retas->proximo = 0;
    while (fscanf(Entrada, "%lf %lf %lf %lf", & ax, & ay, & bx, & by) == 4) {
	if (atual == 0) {
	    atual = retas;
	}
	else {
	    atual = atual->proximo = new reta;
	    if (!atual) {
		fprintf(stderr, "Erro ao alocar memria para leitura de retas: %s\n", strerror(errno));
		return 0;
	    }
	}
	/* Fora as retas a terem orientao de A para B */
	if (ax < bx) {
	  atual->a.x = ax;
	  atual->a.y = ay;
	  atual->b.x = bx;
	  atual->b.y = by;
	}
	else {
	  atual->a.x = bx;
	  atual->a.y = by;
	  atual->b.x = ax;
	  atual->b.y = ay;
	}
	atual->numero_reta = contador;
	atual->proximo = 0;
	
	if (inicio) {
	  inicio = 0;
	  maiory = (atual->a.y > atual->b.y)?atual->a.y:atual->b.y;
	  menory = (atual->a.y < atual->b.y)?atual->a.y:atual->b.y;
	}
	
	if (atual->a.y > maiory) maiory = atual->a.y;
	if (atual->b.y > maiory) maiory = atual->b.y;
	if (atual->a.y < menory) menory = atual->a.y;
	if (atual->b.y < menory) menory = atual->b.y;
	
	contador++;
    }
    return contador;
}


