/****************************************************************************
 * ingenuo.c                                                                *
 * Neste arquivo implementamos o algoritmo ingnuo para o problema de       *
 * interseces de retas em um plano.                                       *
 ****************************************************************************/

#include "funcoes.h"

/* Imprime mensagem explicando o uso */
void uso(char * programa, int saida) {
    fprintf(stderr, "Uso: %s [Nome do arquivo com o conjunto de retas]\n",
	    programa);
    exit(saida);
}

/* Esta funo determina todas as interseces do conjunto de retas */
void interseccao_reta_ingenuo(reta * inicio) {
    reta * j, * k;
    ponto p;
    int passo = 0;
    
    pthread_mutex_lock(& buffer.mutex);
    for (j = inicio; j != 0; j = j->proximo) {
	for (k = j->proximo; k != 0; k = k->proximo) {
	    /* Esperamos a liberao por conta da interface */
	    pthread_cond_wait(& buffer.espera, & buffer.mutex);
	    /* Agora vamos nos divertir */
	    passo++;
	    if (ponto_interseccao_reta(j, k, & p) != 0) {
		insere_interseccao(& p, passo, j, k);
	    }
	    /* Agora vamos atualizar a comparao que foi feita da interface grfica */
	    insere_comparacao_reta(passo, j, k);
	    pthread_cond_signal(& buffer.novo_passo);
	}
    }
    passo++;
    /* Agora precisamos dizer que o algoritmo acabou e matar todas as retas */
    buffer.acabou = 1;
    for (j = inicio; j != 0; j = j->proximo) {
	insere_reta_morta(passo, j);
    }
    pthread_cond_signal(& buffer.novo_passo);
    pthread_mutex_unlock(& buffer.mutex);
}

int main(int argc, char ** argv) {
    FILE * Entrada;
    int numero_retas;
    reta inicio;

    /* Vamos resolver aqui o problema de terminar todas as interseces de um
     * conjunto de retas com O(n^2) */

    glutInit(&argc, argv);

    /* Vamos verificar os argumentos */
    if (argc != 2) {
	fprintf(stderr, "Nmero de argumentos invlido\n");
	uso(argv[0], 1);
    }

    Entrada = fopen(argv[1], "r");
    if (!Entrada) {
        fprintf(stderr, "Erro abrindo arquivo de entrada: \"%s\" %s\n",
                argv[1], strerror(errno));
        exit(3);
    }

    numero_retas = le_pontos(Entrada, & inicio);

    if (numero_retas == 0) {
	fprintf(stderr, "Erro lendo arquivo de entrada, no consegui ler as retas!\n");
	exit(2);
    }

    inicializa_interface(& inicio); /* D uma acertada geral no openGL */

    /* Vamos agora resolver esse problema trivial */
    if(pthread_create(& buffer.algoritmo, 0, (void * (*)(void *)) &
		      interseccao_reta_ingenuo, (void *) & inicio) != 0) {
        printf("No criou thread.\n");
        exit(EXIT_FAILURE);
    }
    
    /* Chama a interface grfica */
    glutMainLoop(); 

    return 0;
}
