/****************************************************************************
 * interface.c                                                              *
 * Neste arquivo implementamos a interface grfica. Ela foi projetada de    *
 * modo a ser independente do algortimo utilizado.                          *
 ****************************************************************************/

/* A idia bsica  usar o openGL para criar a visualizao. A interface  um
 * esquema produtor consumidor. A interface  a consumidora e o algoritmo  o
 * produtor. A interface usa um esquema de lock e buffers para conversar com o
 * algoritmo. No  muito eficiente computacionalmente, mas e da? A grande
 * vantagem dela  que ela permite undo */

/* Estrutura: A interface possui como elementos globais um conjunto de
 * pontos. A cada iterao, a interface dispe de uma lista ligada de
 * interseces para poder desenh-las e uma lista ligada de lista
 * ligada de cores de pontos a cada iterao (Eu ia fazer assim, mas
 * resolvi fazer uma lista de cores para as retas mortas e as retas
 * sendo comparadas. a desvantagem desse modelo  que a lista de retas
 * sendo comparadas precisa ser percorrida sempre, no outro caso, de
 * listas de listas ligadas, eu s percorro o que vou utilizar. Essa
 * abordagem faz com que a interface fique O(n^2) independentemente do
 * algoritmo). Essa lista  percorrida para cada iterao. A interface
 *  orientada a passo. Existe algo do tipo begin passo, end passo,
 * como no openGL e isso define uma mensagem que a interface pega e
 * acerta nas suas variveis e desenha. A informao de undo gasta
 * muita memria, mas quem vai querer visualizar um exemplo com mais
 * de 100 pontos?  O algoritmo fica lento, mesmo com a interface
 * desligada porque mesmo assim ele precisa iteragir com a
 * interface. Isso acontece por causa do undo. */

/* O algoritmo deve dar um lock no mutex antes de comear e avisar que
 * acabou e dar um unlock no mutex depois que terminar o
 * processamento. durante a execuo, ela espera uma condio, quando
 * recebe o sinal, atualiza o passo em questo, e sinaliza a
 * interface. Tudo isso  feito com as variveis do buffer */

#include <GL/glut.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "funcoes.h"

/* Cores que usamos para colorir as retas */
double cor_reta_normal[3] = {1, 1, 1};
double cor_interseccao[3] = {1, .5, .5};
double cor_varredura[3] = {1, 0, 0};
double cor_morta[3] = {.5, .5, .5};
double cor_comparacao1[3] = {0, 1, 0};
double cor_comparacao2[3] = {0, 0, 1};

dados_interprocessos buffer; /* Nossa varivel de comunicao */

/* Temos as variveis de estado: */
reta * retas;

/* Variveis para acertar a exibio na tela */
int resX = 600, resY = 600;
double proporcaoTelaX, proporcaoTelaY;
double proporcaoRetasX, proporcaoRetasY;
double maxX, minX, maxY, minY;

/* Variveis que indicam o passo que estamos */
int passo = 0;
int max_passo = 0;

GLuint desenho_interseccao;

int TipoAnimacao = PASSO; /* Controla o tipo de animao que estamos
			   * utilizando */
unsigned char DirecaoAnimacao = '>'; /* Preciso saber se estou avanando ou
				      * retrocendendo no algoritmo */

int volatile RodandoAnimacao = 0; /* Indica se estamos no meio de uma animao */

interseccao * interseccao_atual = 0; /* Mantm um apontador para o ltimo * elemento da
lista ligada de interseces */



void insere_interseccao(ponto * p, int passo, reta * A, reta * B) {
    if (buffer.i == 0) {
	buffer.i = interseccao_atual = new interseccao;
    }
    else {
	interseccao_atual = interseccao_atual->proximo = new interseccao;
    }

    if (!interseccao_atual) {
	fprintf(stderr, "Erro ao alocar memria para comparao de retas: %s\n", strerror(errno));
	exit(0);
    }
    interseccao_atual->proximo = 0;
    interseccao_atual->A = A;
    interseccao_atual->B = B;
    interseccao_atual->ponto_interseccao.y = p->y;
    interseccao_atual->ponto_interseccao.x = p->x;
    interseccao_atual->passo = passo;
}

coloracao_reta * m_atual = 0; /* Mantm um apontador para o ltimo
			       * elemento da lista ligada de retas mortas */

void insere_reta_morta(int passo, reta * A) {
    if (m_atual == 0) {
	buffer.m = m_atual = new coloracao_reta;
    }
    else {
	m_atual = m_atual->proximo = new coloracao_reta;
    }
    if (!m_atual) {
	fprintf(stderr, "Erro ao alocar memria para comparao de retas: %s\n", strerror(errno));
	exit(0);
    }
    m_atual->r = A;
    m_atual->cor[0] = cor_morta[0];
    m_atual->cor[1] = cor_morta[1];
    m_atual->cor[2] = cor_morta[2];
    m_atual->passo = passo;
    m_atual->proximo = 0;
}

coloracao_reta * varzea_atual = 0; /* Mantm um apontador para o ltimo elemento da
			       * lista ligada de retas sendo comparadas */

void insere_comparacao_reta(int passo, reta * A, reta * B) {
    if (varzea_atual == 0) {
	buffer.c = varzea_atual = new coloracao_reta;
    }
    else {
	varzea_atual = varzea_atual->proximo = new coloracao_reta;
    }
    if (!varzea_atual) {
	fprintf(stderr, "Erro ao alocar memria para comparao de retas: %s\n", strerror(errno));
	exit(0);
    }
    varzea_atual->proximo = new coloracao_reta;
    if (!varzea_atual->proximo) {
	fprintf(stderr, "Erro ao alocar memria para comparao de retas: %s\n", strerror(errno));
	exit(0);
    }
    varzea_atual->cor[0] = cor_comparacao1[0];
    varzea_atual->cor[1] = cor_comparacao1[1];
    varzea_atual->cor[2] = cor_comparacao1[2];
    varzea_atual->r = A;
    varzea_atual->passo = passo;
    varzea_atual = varzea_atual->proximo;

    varzea_atual->r = B;
    varzea_atual->cor[0] = cor_comparacao2[0];
    varzea_atual->cor[1] = cor_comparacao2[1];
    varzea_atual->cor[2] = cor_comparacao2[2];
    varzea_atual->passo = passo;
    varzea_atual->proximo = 0;
}

varredura * v_atual = 0; /* Mantm um apontador para o ltimo elemento da
			       * lista ligada de retas sendo comparadas */
  
void insere_varredura(int passo, double x) {
  if (v_atual == 0) {
      buffer.v = v_atual = new varredura;
  }
  else {
      v_atual = v_atual->proximo = new varredura;
  }
  if (!v_atual) {
      fprintf(stderr, "Erro ao alocar memria para comparao de retas: %s\n", strerror(errno));
      exit(0);
  }
  v_atual->x = x;
  v_atual->passo = passo;
  v_atual->proximo = 0;
}

void desaloca_memoria() {
    reta * atual, * anterior;
    coloracao_reta * c_atual;
    varredura * v_atual;
    interseccao * interseccao_atual;
    interseccao_atual = buffer.i;

    /* Agora vamos desalocar tudo o que alocamos */
    anterior = retas;
    atual = anterior->proximo;
    while (atual != 0) {
	anterior = atual;
	atual = atual->proximo;
	delete anterior;
    }

    interseccao_atual = buffer.i;
    while (interseccao_atual != 0) {
	buffer.i = interseccao_atual;
	interseccao_atual = interseccao_atual->proximo;
	delete buffer.i;
    }

    c_atual = buffer.c;
    while (c_atual != 0) {
	buffer.c = c_atual->proximo;
	delete c_atual;
	c_atual = buffer.c;
    }

    c_atual = buffer.m;
    while (c_atual != 0) {
	buffer.c = c_atual->proximo;
	delete c_atual;
	c_atual = buffer.c;
    }

    v_atual = buffer.v;
    while (v_atual != 0) {
	buffer.v = v_atual->proximo;
	delete v_atual;
	v_atual = buffer.v;
    }
}

/* Antes de sair devemos chamar esta funo para liberar a memria
   usada e imprimir as intersecoes encontradas */
void sair(void) {
    interseccao * interseccao_atual;
    interseccao_atual = buffer.i;

    /* Vamos imprimir as interseces encontradas */

    while (interseccao_atual != 0) {
	printf("Reta %i com %i, (%f, %f)\n", interseccao_atual->A->numero_reta,
	       interseccao_atual->B->numero_reta,
	       interseccao_atual->ponto_interseccao.x,
	       interseccao_atual->ponto_interseccao.y);
	interseccao_atual = interseccao_atual->proximo;
    }
    
    /* Agora vamos desalocar tudo o que alocamos */
    desaloca_memoria();

    /* Acho que j estava na hora */
    exit(0);
}

/* Esta funo faz os desenhos necessrios */
void display(void) { 
    reta * retaT;
    coloracao_reta * corT;
    interseccao * intT;
    varredura * v_atual;

    /* Acertamos a rea de desenho */
    glClearColor(0, 0, 0, 1);
    glClear(GL_COLOR_BUFFER_BIT);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* Arruma o viewport */
    /* Precisamos definir a matriz de transformacao, quero garantir que no
     * haver distoro caso haja um resize */
    glScaled(2/proporcaoTelaX, 2/proporcaoTelaY, 1);
    glTranslated(- proporcaoTelaX / 2 - minX, - proporcaoTelaY / 2 - minY, 0);

    /* Zera a cor dos pixels */
    retaT = retas;
    while (retaT != 0) {
	retaT->cor[0] = cor_reta_normal[0];
	retaT->cor[1] = cor_reta_normal[1];
	retaT->cor[2] = cor_reta_normal[2];
	retaT = retaT->proximo;
    }

    pthread_mutex_lock(& buffer.mutex);

    /* Agora vamos colorir as retas mortas */
    corT = buffer.m;
    while ((corT != 0) && (corT->passo <= passo)) {
	corT->r->cor[0] = corT->cor[0];
	corT->r->cor[1] = corT->cor[1];
	corT->r->cor[2] = corT->cor[2];
	corT = corT->proximo;
    }

    /* Agora vamos colorir as retas sendo comparadas */
    corT = buffer.c;
    while ((corT != 0) && (corT->passo < passo)) {
	corT = corT->proximo;
    }

    while ((corT != 0) && (corT->passo == passo)) {
	corT->r->cor[0] = corT->cor[0];
	corT->r->cor[1] = corT->cor[1];
	corT->r->cor[2] = corT->cor[2];
	corT = corT->proximo;
    }

    /* Linha de varredura */
    v_atual = buffer.v;
    while ((v_atual != 0) && (v_atual->passo < passo))
	v_atual = v_atual->proximo;
    
    glBegin(GL_LINES); {
	while ((v_atual != 0) && (v_atual->passo == passo)) {
	    glColor3f(cor_varredura[0], cor_varredura[1], cor_varredura[2]);
	    glVertex2d(v_atual->x, minX);
	    glVertex2d(v_atual->x, maxX);
	    v_atual = v_atual->proximo;
	}
    } glEnd();

    /* Agora vamos desenhar as interseces */
    intT = buffer.i;
    while ((intT != 0) && (intT->passo <= passo)) {
	glPushMatrix();
	glTranslated(intT->ponto_interseccao.x, intT->ponto_interseccao.y,0);
	glScaled((proporcaoTelaX * 150)/resX,(proporcaoTelaX * 150)/resX,1);
	glCallList(desenho_interseccao);
	glPopMatrix();
	intT = intT->proximo;
    }

    pthread_mutex_unlock(& buffer.mutex);


    /* Vamos desenhar as retas agora */
    glBegin(GL_LINES); {
	retaT = retas;
	
	while (retaT != 0) {
	    glColor3f(retaT->cor[0], retaT->cor[1], retaT->cor[2]);
	    glVertex2d(retaT->a.x, retaT->a.y);
	    glVertex2d(retaT->b.x, retaT->b.y);
	    retaT = retaT->proximo;
	}
    } glEnd();

    glPopMatrix();

    glutSwapBuffers(); 
}

/* Funo de animao */
void AnimaPasso(int tempo) {
    keyboard(DirecaoAnimacao, 0, 0);

    if ((passo <= max_passo) && (passo > 0) && (RodandoAnimacao == 1))
	glutTimerFunc(tempo, AnimaPasso, tempo);
    else
	RodandoAnimacao = 0;
}

/* Trata a entrada do teclado */
void keyboard (unsigned char key, int x, int y) {
    if ((key == 'q') || (key == 'Q')) {
        sair();
    }
    if (key == '>') {
	if (RodandoAnimacao == 0) {
	    RodandoAnimacao = 1;

	    if (TipoAnimacao == SEG1) {
		DirecaoAnimacao = '>';
		glutTimerFunc(1000, AnimaPasso, 1000);
	    }
	    else if (TipoAnimacao == SEG5) {
		DirecaoAnimacao = '>';
		glutTimerFunc(500, AnimaPasso, 500);
	    }
	    else if (TipoAnimacao == SEG25) {
		DirecaoAnimacao = '>';
		glutTimerFunc(25, AnimaPasso, 250);
	    }
	    else if (TipoAnimacao == NO_DELAY) {
		DirecaoAnimacao = '>';
		glutTimerFunc(10, AnimaPasso, 10);
	    }
	    else
		RodandoAnimacao = 0;
	}
	if (TipoAnimacao == NONE) {
	    /* Roda o algoritmo at o fim */
	    while (!buffer.acabou || (passo <= max_passo)) {
		passo++;
		if (passo > max_passo) {
		    if (buffer.acabou)
			passo = max_passo + 1;
		    else {
			pthread_mutex_lock(& buffer.mutex);
			pthread_cond_signal(& buffer.espera);
			pthread_cond_wait(& buffer.novo_passo, &buffer.mutex);
			pthread_mutex_unlock(& buffer.mutex);
			max_passo++;
		    }
		}
	    }
	}
	
	passo++;
	if (passo > max_passo) {
	    if (buffer.acabou)
		passo = max_passo + 1;
	    else {
		pthread_mutex_lock(& buffer.mutex);
		pthread_cond_signal(& buffer.espera);
		pthread_cond_wait(& buffer.novo_passo, &buffer.mutex);
		pthread_mutex_unlock(& buffer.mutex);
		max_passo++;
	    }
	}
	glutPostRedisplay();
    }
    if (key == '<') {
	if (RodandoAnimacao == 0) {
	    RodandoAnimacao = 1;

	    if (TipoAnimacao == SEG1) {
		DirecaoAnimacao = '<';
		glutTimerFunc(1000, AnimaPasso, 1000);
	    }
	    else if (TipoAnimacao == SEG5) {
		DirecaoAnimacao = '<';
		glutTimerFunc(500, AnimaPasso, 500);
	    }
	    else if (TipoAnimacao == SEG25) {
		DirecaoAnimacao = '<';
		glutTimerFunc(250, AnimaPasso, 250);
	    }
	    else if (TipoAnimacao == NO_DELAY) {
		DirecaoAnimacao = '<';
		glutTimerFunc(10, AnimaPasso, 10);
	    }
	    else
		RodandoAnimacao = 0;
	}
	if (TipoAnimacao == NONE) {
	    passo = 1;
	}

	if (passo > 0) {
	    passo--;
	    glutPostRedisplay();
	}
    }
}


void mouse(int btn, int state, int x, int y){
        
}

void reshape(int w, int h) {
    /* No reshape precisamos mapear o tamanho da janela para a proporo que
     * obtemos de forma a simplificar os desenhos */

    resX = w, resY = h;

    glViewport(0,0,w,h);

    /* Calcula as propores de forma que no acontea deformaes */
    if ((proporcaoRetasX / resX) > (proporcaoRetasY / resY)) {
	proporcaoTelaX = proporcaoRetasX;
	proporcaoTelaY = proporcaoTelaX * resY / resX;
    }
    else {
	proporcaoTelaY = proporcaoRetasY;
	proporcaoTelaX = proporcaoTelaY * resX / resY;
    }
}

/* Funo que controla os eventos dos menus */
void choice_selected (int value)
{
    if (value == 1) {
        keyboard('>', 0, 0);
    }
    if (value == 2) {
        keyboard('<', 0, 0);
    }

    if (value == PASSO) {
	RodandoAnimacao = 0;
	TipoAnimacao = PASSO;
    }

    if (value == SEG1) {
	RodandoAnimacao = 0;
	TipoAnimacao = SEG1;
    }

    if (value == SEG5) {
	RodandoAnimacao = 0;
	TipoAnimacao = SEG5;
    }

    if (value == SEG25) {
	RodandoAnimacao = 0;
	TipoAnimacao = SEG25;
    }

    if (value == NO_DELAY) {
	RodandoAnimacao = 0;
	TipoAnimacao = NO_DELAY;
    }

    if (value == NONE) {
	RodandoAnimacao = 0;
	TipoAnimacao = NONE;
    }

    if (value == 3)
        sair();
}

void inicializa_interface(reta * r)
{
    int AnimacaoMenu;
    reta * temp;
    GLUquadricObj * quad;
    retas = r;

    /* Monta a tela da interface e menus */
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutInitWindowSize(resX, resY);
    glutCreateWindow("Projeto de MAC 331 - Deteco de interseco em retas"); 
    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
    glutMouseFunc(mouse);

    AnimacaoMenu = glutCreateMenu(choice_selected);
    glutAddMenuEntry("Passo a Passo", PASSO);
    glutAddMenuEntry("1 segundo", SEG1);
    glutAddMenuEntry(".5 segundos", SEG5);
    glutAddMenuEntry(".25 segundos", SEG25);
    glutAddMenuEntry("sem espera", NO_DELAY);
    glutAddMenuEntry("nenhuma", NONE);

    glutCreateMenu(choice_selected);
    glutAddMenuEntry("Prximo", 1);
    glutAddMenuEntry("Anterior", 2);
    glutAddSubMenu("Animao", AnimacaoMenu);
    glutAddMenuEntry("Sair", 3);
    glutAttachMenu(GLUT_LEFT_BUTTON);

    glutSetCursor(GLUT_CURSOR_CROSSHAIR);

    /* Precisamos achar o maxX, maxY, minX, minY do conjunto de pontos de
     * entrada de forma que possamos visualizar o conjunto completamente */
    temp = retas;
    maxX = minX = temp->a.x;
    maxY = minY = temp->a.y;
    while (temp) {
	if (temp->a.x > maxX)
	    maxX = temp->a.x;
	if (temp->a.x < minX)
	    minX = temp->a.x;
	if (temp->b.x > maxX)
	    maxX = temp->b.x;
	if (temp->b.x < minX)
	    minX = temp->b.x;

	if (temp->a.y > maxY)
	    maxY = temp->a.y;
	if (temp->a.y < minY)
	    minY = temp->a.y;
	if (temp->b.y > maxY)
	    maxY = temp->b.y;
	if (temp->b.y < minY)
	    minY = temp->b.y;
	temp = temp->proximo;
    }

    /* Samos deste loop com uma janela, precisamos agora dar uma
       "folguinha" para esses valores */
    maxX += (maxX - minX) * .1;
    maxY += (maxY - minY) * .1;
    minX -= (maxX - minX) * .1;
    minY -= (maxY - minY) * .1;
    proporcaoRetasX = (maxX - minX);
    proporcaoRetasY = (maxY - minY);

    /* Inicializa a nossa varivel de comunicao */
    buffer.i = 0;
    buffer.c = 0;
    buffer.m = 0;
    buffer.v = 0;
    buffer.acabou = 0;
    pthread_cond_init(& buffer.espera, 0);
    pthread_cond_init(& buffer.novo_passo, 0);
    pthread_mutex_init(& buffer.mutex, 0);

    /* Gera o desenho da interseco */
    desenho_interseccao = glGenLists(1);
    quad = gluNewQuadric();
    gluQuadricDrawStyle(quad, (GLenum) GLU_LINE);
    glNewList(desenho_interseccao, GL_COMPILE); {
	glColor3f(cor_interseccao[0], cor_interseccao[1], cor_interseccao[2]);
	gluDisk(quad, .05, .05, 20, 1);
	glBegin(GL_LINES); {
	    glVertex2d(0, .05);
	    glVertex2d(0, -.05);
	    glVertex2d(.05, 0);
	    glVertex2d(-.05, 0);
	} glEnd();

    } glEndList();
}
