#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "avl.h"
#include "event_queue.h"
#include "sweep_line.h"
#include "funcoes.h"
#include "interface.h"
#define _REENTRANT


void TrataPontoEvento(event_queue *, sweep_line *, ponto_evento*);
void EncontreNovosEventos(event_queue *, celula_pt, celula_pt, ponto_evento*);

extern double maiory, menory;
int passo_varredura;

/* Imprime mensagem explicando o uso */
void uso(char * programa, int saida) {
    fprintf(stderr, "Uso: %s [Nome do arquivo com o conjunto de segmentos]\n",
	    programa);
    exit(saida);
}

typedef struct a {
    event_queue * Q;
    sweep_line * T;
} args;

void resolve_varredura(args * a) {
    event_queue * Q;
    sweep_line * T;
    Q = a->Q;
    T = a->T;

    pthread_mutex_lock(& buffer.mutex);
    printf("->Passo %i\n", passo_varredura);
    while (!Q->vazia()) {
	pthread_cond_wait(& buffer.espera, & buffer.mutex);
	passo_varredura++;
	printf("->Passo %i\n", passo_varredura);
	TrataPontoEvento(Q, T, Q->minimo());
	printf("Tratou ponto evento %d\n", passo_varredura);
	pthread_cond_signal(& buffer.novo_passo);
    }
    buffer.acabou = 1;
    pthread_mutex_unlock(& buffer.mutex);
}

int main(int argc, char **argv) {
  event_queue Q;
  sweep_line T;
  unsigned int i, num_retas;
  reta inicio, *atual;
  FILE *arq;
  ponto_evento *pev_aux;
  args a;

  glutInit(&argc, argv);
 
  /* Vamos verificar os argumentos */
  if (argc != 2) {
    fprintf(stderr, "Nmero de argumentos invlido\n");
    uso(argv[0], 1);
  }

  arq = fopen(argv[1], "r");
  if (!arq) {
    fprintf(stderr, "Erro abrindo arquivo de entrada: \"%s\" %s\n", argv[1], strerror(errno));
    exit(3);
  }

  num_retas = le_pontos(arq, &inicio);

  if (num_retas == 0) {
    fprintf(stderr, "Erro lendo arquivo de entrada, no consegui ler os segmentos!\n");
    exit(2);
  }

  inicializa_interface(& inicio); /* D uma acertada geral no openGL */

  /* Inicilaiza a fila do Algoritmo */
  // Insere os pontos extremos dos segmentos em Q
  atual = &inicio;
  for (i = 0; i < num_retas; i++, atual = atual->proximo) {
    celula_pt resultado; 

    // ponto extremo esquerdo
    pev_aux = new ponto_evento(atual->a, ESQ, atual);

    if ((resultado = Q.busca_exata(pev_aux))) {
      resultado->i.p->adiciona_pes(ESQ, atual);
      insere_interseccao(&resultado->i.p->a, 0, atual, resultado->i.p->pseg->s);
    }
    else
      Q.insert(pev_aux);
  
    // ponto extremo direito
    pev_aux = new ponto_evento(atual->b, DIR, atual);

    if ((resultado = Q.busca_exata(pev_aux))) {
      insere_interseccao(&resultado->i.p->a, 0, atual, resultado->i.p->pseg->s);
      resultado->i.p->adiciona_pes(DIR, atual);
    }
    else
      Q.insert(pev_aux);
  }


  
  // algoritmo principal
  a.T = &T;
  a.Q = &Q;

  printf("(%p, %p)\n", &T, &Q); 
//  printf("(%p, %p)\n", args[0], args[1]); 

  if(pthread_create(& buffer.algoritmo, 0, (void * (*)(void *)) &
		    resolve_varredura, (void *) & a) != 0) {
      fprintf(stderr, "No criou thread.\n");
      exit(EXIT_FAILURE);
  }


//  resolve_varredura(& a);
  glutMainLoop(); 
  
  
  return 0;
}

// NO TRATA SEGMENTO VERTICAL
void TrataPontoEvento(event_queue *Q, sweep_line *T, ponto_evento* p) {
  sweep_line Taux;
  // Flag para verificar se o conjunto E(p) unio C(p) no est vazio
  unsigned short esq_centro_nao_vazio = 0;
  // Flag de interseccao
  unsigned short tem_interseccao = 0;
  // Auxiliares
  ponto_evento_segmento* pseg_aux;
  celula_pt dir, esq;

  dir = esq = NULL;

  // Preparativos para uso da Reta Vertical
  reta vertical;
  ponto pto;
  int cont;
  double soma;
  
  cont = 0;
  soma = 0;
  
  vertical.a.y = menory;
  vertical.b.y = maiory;
  // Fim Reta Vertical

  /* Vamos desenhar a linha de varredura */
  insere_varredura(passo_varredura, p->a.x);

  printf("+ tratando ponto_evento: (%.1f, %.1f)\n", p->a.x, p->a.y);
  
  if (p->ha_interseccao()) {
    tem_interseccao = 1;
    printf("\n*** INTERSECO ENTRE SEGMENTOS ***\n");
    printf("--> ponto de interseco: (%.1f, %.1f)\n", p->a.x, p->a.y);

    
    for (pseg_aux = p->pseg; pseg_aux; pseg_aux = pseg_aux->prox)
      printf("--> segmento: (%.1f,%.1f), (%.1f, %.1f)\n", pseg_aux->s->a.x, pseg_aux->s->a.y, pseg_aux->s->b.x, pseg_aux->s->b.y);
    printf("\n");

  }

#ifdef DEBUB
  printf("Antes de Remover\n");
  T->imprimeArvore();
#endif DEBUG

  // ajusta reta vertical para a remoo dos segmentos
  vertical.a.x = vertical.b.x = p->a.x - INCREMENTO_PROCURA;

  // remover todos os segmentos cujo tipo_ponto == DIR ou CENTRO (D(p) unio C(p))
  for (pseg_aux = p->pseg; pseg_aux; pseg_aux = pseg_aux->prox)

    // todos os segmentos cujo tipo_ponto == DIR so segmentos MORTOS 
    if (pseg_aux->t == DIR || pseg_aux->t == CENTRO) {
      celula_pt resultado;
      ponto_evento* pev_temp;

      if (tem_interseccao) {
	ponto_interseccao_reta(&vertical, pseg_aux->s, &pto);
	ponto_evento pev_interseccao_vertical(pto);
	pev_temp = &pev_interseccao_vertical;
      }
      else
	pev_temp = p;

      // guarda-se os segmentos  direita e  esquerda de p, para eventual uso
      if (!pseg_aux->prox) {
	resultado = T->busca(pseg_aux->s, pev_temp);
	
	// teste de sanidade
	if (!resultado) { 
	  //T->imprime_arvore();
	  fprintf(stderr, "ERRO - TrataPontoEvento: buscaSEG() para remoo.\n"); 
	  exit(1); 
	}

	if (resultado->esq) esq = resultado->esq;
	else if (resultado->pai && resultado->pai->dir == resultado) esq = resultado->pai;
	else esq = NULL;
	
	if (resultado->dir) dir = resultado->dir;
	else if (resultado->pai  && resultado->pai->esq == resultado) dir = resultado->pai;
	else dir = NULL;
      }

      // e finalmente remove o segmento que o ponto evento esteja  direita ou no centro
      T->remove(pseg_aux->s, pev_temp);
      /* Atualizamos o conjunto de retas mortas da interface */
      
      
      if (pseg_aux->t == DIR) {
	  insere_reta_morta(passo_varredura, pseg_aux->s);
      }
      

#ifdef DEBUG
      printf("Removeu (ou no)\n");
      T->imprime_arvore();
#endif DEBUG
    }

  // ajusta reta vertical
  vertical.a.x = vertical.b.x = p->a.x + INCREMENTO_PROCURA;

  // inserir os segmentos em E(p) unio C(p)
  for (pseg_aux = p->pseg; pseg_aux; pseg_aux = pseg_aux->prox)
    if (pseg_aux->t == ESQ || pseg_aux->t == CENTRO) {
      ponto_evento* pev_temp;

      // arruma a flag, avisando que o conjunto E(p) U C(p) no est vazio
      if (!esq_centro_nao_vazio) esq_centro_nao_vazio = 1;
      
      if (tem_interseccao) {
	pev_temp = new ponto_evento(p);
	pev_temp->a.x += INCREMENTO_PROCURA;
      }
      else
	pev_temp = p;      

      T->insert(pseg_aux->s, pev_temp, tem_interseccao);
      Taux.insert(pseg_aux->s, pev_temp, tem_interseccao);
      
#ifdef DEBUG
      printf("Inseriu\n");
      T->imprime_arvore();
#endif DEBUG
    }

  // encontrar novos segmentos
  if (!esq_centro_nao_vazio) {
    if (dir && esq) EncontreNovosEventos(Q, esq, dir, p);
  }
  else {
    celula_pt resultado, mais_esquerda, mais_direita;
    ponto_evento *pev_temp, *pev_interseccao_vertical;
    
    // segmento mais  esquerda na rvore
    mais_esquerda = Taux.avl_minimo(); 

    // tratamento diferenciado para interseccao. utiliza-se a interseccao com a reta vertical.
    if (tem_interseccao) {
      ponto_interseccao_reta(&vertical, reta(mais_esquerda), &pto);
      // ltimos parmetros so irrelevantes
      pev_interseccao_vertical = new ponto_evento(pto);
      pev_temp = pev_interseccao_vertical;
    }
    else
      pev_temp = p;

#ifdef DEBUG
    T->imprime_arvore();
    printf("TAUX\n");
    Taux.imprime_arvore();
    printf("MIN - Interseccao da RETA VERTICAL: (%.2f, %.2f)\n", pto.x, pto.y);
    printf("Com reta: (%.1f,%.1f), (%.1f, %.1f)\n", reta(mais_esquerda)->a.x, reta(mais_esquerda)->a.y, reta(mais_esquerda)->b.x, reta(mais_esquerda)->b.y);
    if (Collinear(&(reta(mais_esquerda)->a), &(reta(mais_esquerda)->b), &(pto))) printf("COLINEAR!\n");
#endif DEBUG

    resultado = T->busca(reta(mais_esquerda), pev_temp);

    // teste de sanidade
    if (!resultado) { fprintf(stderr, "TrataPontoEvento: ERRO - buscaSEG() para EncontrarNovosEventos.\n"); exit(1); }

    // segmento predecessor
    if (!resultado->pai || resultado->esq)
      resultado = resultado->esq;
    else if (resultado->pai->dir == resultado)
      resultado = resultado->pai;
    else if (resultado->pai->pai && resultado->pai->pai->dir == resultado->pai)
      resultado = resultado->pai->pai;
    else
      resultado = NULL;
    
    // se existir o segmento predecessor, chama o EncontreNovosEventos
    if (resultado) EncontreNovosEventos(Q, resultado, mais_esquerda, p);
    
    // segmento mais  direita na rvore
    mais_direita = Taux.avl_maximo(); 
    
    // tratamento diferenciado para interseccao. utiliza-se a interseccao com a reta vertical.
    if (tem_interseccao) {
      ponto_interseccao_reta(&vertical, reta(mais_direita), &pto);
      pev_interseccao_vertical->a = pto;
      pev_temp = pev_interseccao_vertical;
    }
    else
      pev_temp = p;

#ifdef DEBUG
    printf("MAX - Interseccao da RETA VERTICAL: (%.2f, %.2f)\n", pto.x, pto.y);
    printf("Com reta: (%.1f,%.1f), (%.1f, %.1f)\n", reta(mais_direita)->a.x, reta(mais_direita)->a.y, reta(mais_direita)->b.x, reta(mais_direita)->b.y);
#endif DEBUG    

    resultado = T->busca(reta(mais_direita), pev_temp);
    if (!resultado) { fprintf(stderr, "TrataPontoEvento: ERRO - buscaSEG() para EncontrarNovosEventos.\n"); exit(1); }

    // segmento sucessor
    if (!resultado->pai || resultado->dir)
      resultado = resultado->dir;
    else if (resultado->pai->dir != resultado)
      resultado = resultado->pai;
    else if (resultado->pai->pai && resultado->pai->pai->esq == resultado->pai)
      resultado = resultado->pai->pai;
    else
      resultado = NULL;

    // se existir o segmento sucessor, chama o EncontreNovosEventos
    if (resultado) EncontreNovosEventos(Q, mais_direita, resultado, p);    
  }  
    
}

void EncontreNovosEventos(event_queue *Q, celula_pt esq, celula_pt dir, ponto_evento *p) {
  ponto pto;
  printf("### testando interseco entre:\n");
  printf("--> segmento: (%.1f,%.1f), (%.1f, %.1f)\n", reta(esq)->a.x, reta(esq)->a.y, reta(esq)->b.x, reta(esq)->b.y);
  printf("--> segmento: (%.1f,%.1f), (%.1f, %.1f)\n", reta(dir)->a.x, reta(dir)->a.y, reta(dir)->b.x, reta(dir)->b.y);

  insere_comparacao_reta(passo_varredura, reta(esq), reta(dir));

  if (ponto_interseccao_reta(reta(esq), reta(dir), &pto) && (pto.x > p->a.x)) {
    tipo_ponto tp_temp1, tp_temp2;
    ponto_evento* pev_aux;
    celula_pt resultado;

    insere_interseccao(&pto, passo_varredura, reta(esq), reta(dir));

    if (pto.x == reta(esq)->a.x && pto.y == reta(esq)->a.y) tp_temp1 = ESQ;
    else if (pto.x == reta(esq)->b.x && pto.y == reta(esq)->b.y) tp_temp1 = DIR;
    else tp_temp1 = CENTRO;    

    // verifica se a interseco est na direita, na esquerda ou no centro do segmento
    if (pto.x == reta(dir)->a.x && pto.y == reta(dir)->a.y) tp_temp2 = ESQ;
    else if (pto.x == reta(dir)->b.x && pto.y == reta(dir)->b.y) tp_temp2 = DIR;
    else tp_temp2 = CENTRO;

    // talvez de para arrumar esse cdigo. est meio FEIO
    if (tp_temp1 == CENTRO) {
      pev_aux = new ponto_evento(pto, tp_temp1, reta(esq));

      if (tp_temp2 == CENTRO)
	pev_aux->adiciona_pes(tp_temp2, reta(dir));
    }
    else if (tp_temp2 == CENTRO) {
      pev_aux = new ponto_evento(pto, CENTRO, reta(dir));
      if ((resultado = Q->busca_exata(pev_aux))) {
	printf("$ verificou se os segs no pev existente\n");
	Q->verifica_interseccao(resultado->i.p, pev_aux);
      }
      else {
	fprintf(stderr, "ERRO - EncontreNovosEventos: Q->busca_exataSEG().\n"); 
	exit(1);
      }
    }

    if (tp_temp1 == CENTRO && tp_temp2 == CENTRO) 
      if ((resultado = Q->busca_exata(pev_aux))) {
	printf("$ verificou se os segs no pev existente\n");
	Q->verifica_interseccao(resultado->i.p, pev_aux);
      }
      else {
	Q->insert(pev_aux);      
	printf("=== inseriu interseco: (%.1f, %.1f)\n", pev_aux->a.x, pev_aux->a.y);
      }
  }
}
